// power_directToMemoryEmitter.C

#include "power_directToMemoryEmitter.h"
#include "power_instr.h"
#include "power64_abi.h"
#include "kernelDriver.h"

extern void flush_a_little(kptr_t); // .s file

// The following isn't static because various launchers use it.  In fact it's
// so general that it should probably be moved.
void flush_a_bunch(kptr_t iaddr, unsigned numbytes) {
    //We can't flush kernel addresses from userspace, so must ioctl.
   extern kernelDriver *global_kernelDriver; // sorry about this ugliness
   global_kernelDriver->flushIcacheRange(iaddr, numbytes);

   //OLD CODE
    // flush requires 4-byte alignment.  We don't enforce this (by nand'ing with 0x3)
   // because we know that we're already 4-byte aligned.
   //It is unclear what the instruction cache block size is on Power processors, and 
   //whether it varies.  Let's be safe for now, and leave it as 4
  
  
   
// register kptr_t addr = iaddr;
//   while (numbytes) {
//      flush_a_little(addr); // asm
//      asm("icbi 0, %0" : /* no outputs */ : "r" (addr));
//      addr += 4;
//      numbytes -= 4;
//   } 
   //should there be an isync here to make sure the pipeline is flushed??
}

void power_directToMemoryEmitter::complete() {
   // We just wrote a bunch of instructions.  They just made it to the data cache.
   // We need to issue cache block invalidate instructions to make sure they reach
   // the icache
   
   assert(insnBytesEmitted <= totalNumInsnBytesOKToWrite);
   //flush_a_bunch(whereEmittingStartsInKerninstd, insnBytesEmitted);
   //We flush the icache for the memory we write in kernel space, not
   //kerninstd space.
}

void power_directToMemoryEmitter::emitSave(const unsigned numRegs) {
  assert(numRegs <= 19);
  int regOffset =  ((power64_abi &)theABI).getGeneralRegSaveFrameOffset();
  const unsigned frameSize = theABI.getBigFrameAlignedNumBytes();
  //First, we need to generate a stack frame.
  instr_t *instr = new power_instr(power_instr::store, 
				   power_instr::stDoubleUpdateDS, 
				   power_reg::gpr1, 
				   power_reg::gpr1, //stack pointer
				   -frameSize); 
  
  emit(instr);
  
  //store registers from r31 down on the stack
  for (unsigned i = 0; i < numRegs ; i++) {
    power_reg regToSave =  power_reg(power_reg::rawIntReg, 31-i); 
    instr = new power_instr(power_instr::store, power_instr::stDoubleDS, 
				     regToSave,  
				     power_reg::gpr1, //stack pointer  
				     regOffset+GPRSIZE*i //offset off SP 
		        );
    emit(instr);
  }  
}

void power_directToMemoryEmitter::emitRestore(const regset_t *regsToRestore) {
  assert(regsToRestore->count() <= 19);
  int regOffset =  ((power64_abi &)theABI).getGeneralRegSaveFrameOffset();
  const unsigned frameSize = theABI.getBigFrameAlignedNumBytes();

  //we assume the registers to restore are nonvolatile regs starting with GPR31
  //and consecutively numbered registers backwards.
  for (unsigned i = 0; i < regsToRestore->count() ; i++) {
    power_reg regToRestore =  power_reg(power_reg::rawIntReg, 31-i); 
    instr_t *instr = new power_instr(power_instr::load, power_instr::ldDoubleDS, 
				     regToRestore,  
				     power_reg::gpr1, //stack pointer  
				     regOffset + GPRSIZE*i //offset off SP 
	);
    emit(instr);
    assert(regsToRestore->exists(regToRestore));
  }

  //restore the stack pointer
  instr_t *instr = new power_instr(power_instr::load, 
				   power_instr::ldDoubleX, 
				   power_reg::gpr1, 
				   power_reg::gpr0, //no second addr reg.
				   power_reg::gpr1); 
  
 
  emit(instr);
}
