#include "primitive.h"
#include "dataHeap.h"
#include "kapi.h"

extern kapi_manager kmgr;

// Compute log2(size)
unsigned primitive::logSize(unsigned size) const
{
    assert(size == 64 && "Primitive of unsupported size");
    return 6;
}

// Return a snippet that computes (cpu_id * cpu_stride)
kapi_arith_expr primitive::computeCpuOffset() const
{
    unsigned cpu_stride = getStrideSize();
    unsigned log2_cpu_stride = logSize(cpu_stride);

    return kapi_arith_expr(kapi_shift_left, kapi_cpuid_expr(),
			   kapi_const_expr(log2_cpu_stride));
}


// Wrap the given code with lock/unlock calls that prevent it from
// being preempted by raising PIL of the current CPU
kapi_snippet
primitive::preemptionProtectedCode(const kapi_snippet &code) const
{
   kapi_vector<kapi_snippet> lock_code_unlock;

#ifdef sparc_sun_solaris2_7

   const unsigned disp_level = kmgr.getDispatcherLevel();
    
   kapi_getpil_expr orig_level;
   kapi_bool_expr is_below_disp_level(kapi_lt, orig_level,
                                      kapi_const_expr(disp_level));
   kapi_if_expr raise_level_if_low(is_below_disp_level, 
                                   kapi_setpil_expr(kapi_const_expr(disp_level)));

   kapi_setpil_expr restore_level(orig_level);

   lock_code_unlock.push_back(orig_level);
   lock_code_unlock.push_back(raise_level_if_low);
   lock_code_unlock.push_back(code);
   lock_code_unlock.push_back(restore_level);
    
#elif defined(i386_unknown_linux2_4) || defined(ppc64_unknown_linux2_4)

   kapi_disableint_expr clear_local_interrupts;
   kapi_enableint_expr set_local_interrupts;
   
   lock_code_unlock.push_back(clear_local_interrupts);
   lock_code_unlock.push_back(code);
   lock_code_unlock.push_back(set_local_interrupts);
    
#endif

   return kapi_sequence_expr(lock_code_unlock);
}




